/* extension.js
 *
 * Project: Exchange Electricity Price Indicator
 * Website: https://codeberg.org/artnay/exchange-electricity-price-indicator
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

import St from 'gi://St';
import GObject from 'gi://GObject';
import GLib from 'gi://GLib';
import Soup from 'gi://Soup';
import Clutter from 'gi://Clutter';
import Gio from 'gi://Gio';
import * as Main from 'resource:///org/gnome/shell/ui/main.js';
import * as PanelMenu from 'resource:///org/gnome/shell/ui/panelMenu.js';
import * as PopupMenu from 'resource:///org/gnome/shell/ui/popupMenu.js';
import {
    Extension as ExtensionBase,
    gettext as _,
} from 'resource:///org/gnome/shell/extensions/extension.js';

const API_URL = 'https://api.porssisahko.net/v1/latest-prices.json';
const CACHE_FILENAME = 'eepi-latest-prices-cache.json';
const PRICE_UPDATE_HOUR = 14;
const PRICE_UPDATE_MINUTE = 20;
const HELSINKI_TIMEZONE = 'Europe/Helsinki';

const ElectricityPriceIndicator = GObject.registerClass(
    class ElectricityPriceIndicator extends PanelMenu.Button {
        _init() {
            super._init(0.0, _('Exchange Electricity Price Indicator'));

            // Initialize cache
            this._cacheFile = Gio.File.new_for_path(
                GLib.build_filenamev([
                    GLib.get_user_cache_dir(),
                    'gnome-shell',
                    CACHE_FILENAME,
                ])
            );
            this._cachedData = null;
            this._lastUpdateTime = null;

            // Store signal handlers
            this._signalHandlers = [];

            // Create UI components
            this._createUI();
            this._createMenu();

            // Initial refresh
            this._refresh();

            // Set up hourly refresh 5 seconds after the hour
            const scheduleNextRefresh = () => {
                const now = new Date();
                const nextHour = new Date(now);
                nextHour.setHours(now.getHours() + 1, 0, 5, 0);
                const delay = Math.floor((nextHour - now) / 1000);

                this._refreshTimeout = GLib.timeout_add_seconds(
                    GLib.PRIORITY_DEFAULT,
                    delay,
                    () => {
                        this._refresh();
                        scheduleNextRefresh();
                        return GLib.SOURCE_REMOVE;
                    }
                );
            };
            scheduleNextRefresh();
        }

        _createUI() {
            // Create panel display
            const hbox = new St.BoxLayout({
                style_class: 'panel-status-menu-box',
            });
            this.icon = new St.Icon({
                icon_name: 'dialog-information-symbolic',
                style_class: 'system-status-icon',
            });
            this.priceLabel = new St.Label({
                text: _('Loading…'),
                y_align: Clutter.ActorAlign.CENTER,
            });

            hbox.add_child(this.icon);
            hbox.add_child(this.priceLabel);
            this.add_child(hbox);
        }

        _getHelsinkiTime() {
            // Get current time in UTC
            const now = new Date();

            // Create a formatter for Helsinki time
            const formatter = new Intl.DateTimeFormat('en-US', {
                timeZone: HELSINKI_TIMEZONE,
                hour: 'numeric',
                minute: 'numeric',
                second: 'numeric',
                hour12: false,
            });

            // Parse the Helsinki time components
            const helsinkiTimeStr = formatter.format(now);
            const [hours, minutes] = helsinkiTimeStr.split(':').map(Number);
            // const [hours, minutes, seconds] = helsinkiTimeStr.split(':').map(Number);

            // Create a new date object with Helsinki time components
            const helsinkiTime = new Date(now);
            helsinkiTime.setHours(hours, minutes);
            //  helsinkiTime.setHours(hours, minutes, seconds);

            return helsinkiTime;
        }

        _getNextUpdateTime() {
            const helsinkiTime = this._getHelsinkiTime();
            const nextUpdate = new Date(helsinkiTime);

            // Set to today's update time
            nextUpdate.setHours(PRICE_UPDATE_HOUR, PRICE_UPDATE_MINUTE, 0, 0);

            // If we're past today's update time, set to tomorrow's update time
            if (helsinkiTime > nextUpdate)
                nextUpdate.setDate(nextUpdate.getDate() + 1);

            return nextUpdate;
        }

        _shouldFetchNewData() {
            const helsinkiTime = this._getHelsinkiTime();

            // If we don't have cached data, we should fetch
            if (
                !this._cachedData ||
                !this._cachedData.prices ||
                !Array.isArray(this._cachedData.prices)
            )
                return true;

            // Get today and tomorrow dates in Helsinki timezone
            const today = new Date(helsinkiTime);
            const tomorrow = new Date(helsinkiTime);
            tomorrow.setDate(tomorrow.getDate() + 1);

            // Set to start of day
            today.setHours(0, 0, 0, 0);
            tomorrow.setHours(0, 0, 0, 0);

            // Find current price
            const currentPrice = this._cachedData.prices.find(price => {
                const startDate = new Date(price.startDate);
                const endDate = new Date(price.endDate);
                return startDate <= helsinkiTime && endDate > helsinkiTime;
            });

            // If no current price, we need to fetch
            if (!currentPrice) return true;

            // Check if we have complete data for today and tomorrow
            const todayData = this._cachedData.prices.filter(price => {
                const priceDate = new Date(price.startDate);
                return priceDate >= today && priceDate < tomorrow;
            });

            const tomorrowData = this._cachedData.prices.filter(price => {
                const priceDate = new Date(price.startDate);
                const nextDay = new Date(tomorrow);
                nextDay.setDate(nextDay.getDate() + 1);
                return priceDate >= tomorrow && priceDate < nextDay;
            });

            // Check if we have all 24 hourly prices for each day
            if (todayData.length < 24 || tomorrowData.length < 24) return true;

            // Check if we need to fetch based on the daily update schedule
            const updateTime = new Date(this._cachedData.timestamp);
            const nextUpdateTime = this._getNextUpdateTime();

            return (
                updateTime < nextUpdateTime && helsinkiTime >= nextUpdateTime
            );
        }

        async _fetchDataWithRetry(maxRetries = 5, retryDelay = 60) {
            let attempt = 1;

            while (attempt <= maxRetries) {
                try {
                    const session = new Soup.Session();
                    const message = Soup.Message.new('GET', API_URL);

                    const bytes = await new Promise((resolve, reject) => {
                        session.send_and_read_async(
                            message,
                            GLib.PRIORITY_DEFAULT,
                            null,
                            (session, result) => {
                                try {
                                    if (
                                        message.get_status() !== Soup.Status.OK
                                    ) {
                                        throw new Error(
                                            `HTTP Error: ${message.get_status()}`
                                        );
                                    }

                                    resolve(
                                        session.send_and_read_finish(result)
                                    );
                                } catch (error) {
                                    reject(error);
                                }
                            }
                        );
                    });

                    const decoder = new TextDecoder('utf-8');
                    const responseText = decoder.decode(bytes.get_data());
                    const response = JSON.parse(responseText);

                    if (!response.prices || !Array.isArray(response.prices))
                        throw new Error('Invalid API response format');

                    return response.prices;
                } catch (error) {
                    if (attempt === maxRetries) {
                        logError(
                            `Final attempt ${attempt}/${maxRetries} failed: ${error}`
                        );
                        throw new Error(
                            `Failed to fetch data after ${maxRetries} attempts: ${error.message}`
                        );
                    }

                    logError(
                        `Attempt ${attempt}/${maxRetries} failed: ${error}. Retrying in ${retryDelay} seconds...`
                    );

                    // Wait for the specified delay before retrying
                    await new Promise(resolve =>
                        GLib.timeout_add_seconds(
                            GLib.PRIORITY_DEFAULT,
                            retryDelay,
                            () => {
                                resolve();
                                return GLib.SOURCE_REMOVE;
                            }
                        )
                    );

                    attempt++;
                }
            }
        }

        async _loadCache() {
            try {
                if (!this._cacheFile.query_exists(null)) return null;

                const [success, contents] = await new Promise(resolve => {
                    this._cacheFile.load_contents_async(
                        null,
                        (file, result) => {
                            try {
                                const [success, contents] =
                                    file.load_contents_finish(result);
                                resolve([success, contents]);
                                // } catch (_e) {
                            } catch {
                                resolve([false, null]);
                            }
                        }
                    );
                });

                if (!success || !contents) return null;

                const decoder = new TextDecoder('utf-8');
                const cacheData = JSON.parse(decoder.decode(contents));

                if (!cacheData.timestamp || !cacheData.prices) return null;

                return cacheData;
            } catch (error) {
                console.error('Error loading cache:', error);
                return null;
            }
        }

        async _saveCache(newData) {
            try {
                // Create cache directory if it doesn't exist
                const cacheDir = Gio.File.new_for_path(
                    GLib.build_filenamev([
                        GLib.get_user_cache_dir(),
                        'gnome-shell',
                    ])
                );
                if (!cacheDir.query_exists(null))
                    cacheDir.make_directory_with_parents(null);

                // Load existing cache data
                let existingData = await this._loadCache();
                if (!existingData) existingData = {prices: []};

                // Create a map of existing prices by startDate for faster lookup
                const existingPricesMap = new Map(
                    existingData.prices.map(price => [price.startDate, price])
                );

                // Merge new data with existing data, replacing duplicates
                newData.forEach(price => {
                    existingPricesMap.set(price.startDate, price);
                });

                // Convert map back to array and sort by startDate
                const mergedPrices = Array.from(
                    existingPricesMap.values()
                ).sort((a, b) => new Date(a.startDate) - new Date(b.startDate));

                // Remove prices older than 7 days to keep cache size manageable
                const sevenDaysAgo = new Date();
                sevenDaysAgo.setDate(sevenDaysAgo.getDate() - 7);
                const filteredPrices = mergedPrices.filter(
                    price => new Date(price.startDate) >= sevenDaysAgo
                );

                // Save cache data with Helsinki timestamp
                const cacheData = {
                    timestamp: this._getHelsinkiTime().toISOString(),
                    prices: filteredPrices,
                };

                const encoder = new TextEncoder();
                const contents = encoder.encode(JSON.stringify(cacheData));

                await new Promise((resolve, reject) => {
                    this._cacheFile.replace_contents_bytes_async(
                        new GLib.Bytes(contents),
                        null,
                        false,
                        Gio.FileCreateFlags.NONE,
                        null,
                        (file, result) => {
                            try {
                                file.replace_contents_finish(result);
                                resolve();
                            } catch (e) {
                                reject(e);
                            }
                        }
                    );
                });

                // Update the in-memory cache
                this._cachedData = cacheData;
            } catch (error) {
                logError('Error saving cache:', error);
                throw error;
            }
        }

        _formatDate(date) {
            // Convert JavaScript Date to GLib.DateTime
            const gDateTime = GLib.DateTime.new(
                GLib.TimeZone.new(HELSINKI_TIMEZONE),
                date.getFullYear(),
                date.getMonth() + 1, // JavaScript months are 0-based
                date.getDate(),
                date.getHours(),
                date.getMinutes(),
                date.getSeconds()
            );

            // Format using strftime - Translators: modify the date format string - Use %x for locale's preferred date representation
            return gDateTime.format(_('%x')); // Use %x for locale's preferred date representation
        }

        _formatTime(date) {
            const gDateTime = GLib.DateTime.new(
                GLib.TimeZone.new(HELSINKI_TIMEZONE),
                date.getFullYear(),
                date.getMonth() + 1,
                date.getDate(),
                date.getHours(),
                date.getMinutes(),
                date.getSeconds()
            );

            // Format using strftime - Translators: modify the time format string - %R represents time in 24-hour notation (HH:MM)
            return gDateTime.format(_('%R')); // %R represents time in 24-hour notation (HH:MM)
        }

        _formatDateTime(dateString) {
            const date = new Date(dateString);
            return date.toLocaleString(undefined, {
                timeZone: HELSINKI_TIMEZONE,
                //    year: 'numeric',
                //    month: '2-digit',
                //    day: '2-digit',
                hour: '2-digit',
                minute: '2-digit',
                hour12: false,
            });
        }

        _formatHourRange(date) {
            return date.toLocaleTimeString(undefined, {
                hour: '2-digit',
                minute: '2-digit',
                timeZone: HELSINKI_TIMEZONE,
            });
        }

        _createMenu() {
            // State for view management
            this._currentView = 'today'; // 'today' or 'tomorrow'

            // Add header
            const headerItem = new PopupMenu.PopupBaseMenuItem({
                reactive: false,
                style_class: 'header-item',
            });

            const headerLabel = new St.Label({
                text: _('⚡ Exchange Electricity Prices'),
                style_class: 'header-label',
                x_expand: true,
                y_align: Clutter.ActorAlign.CENTER,
                x_align: Clutter.ActorAlign.CENTER,
            });

            headerItem.actor.add_child(headerLabel);
            this.menu.addMenuItem(headerItem);
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

            // Create navigation buttons
            const navigationItem = new PopupMenu.PopupBaseMenuItem({
                reactive: false,
                style_class: 'navigation-item',
            });

            const navigationBox = new St.BoxLayout({
                style_class: 'navigation-box',
                x_expand: true,
                height: 22, // Set fixed height for the navigation box
            });

            // Left arrow button
            this.prevButton = new St.Button({
                reactive: true,
                can_focus: true,
                track_hover: true,
                style_class: 'navigation-button',
                width: 22,
                height: 22,
                child: new St.Icon({
                    icon_name: 'go-previous-symbolic',
                    icon_size: 14,
                    style_class: 'navigation-icon',
                }),
            });

            // View label
            this.viewLabel = new St.Label({
                text: _('Today'),
                style_class: 'view-label',
                x_expand: true,
                y_align: Clutter.ActorAlign.CENTER,
                x_align: Clutter.ActorAlign.CENTER,
            });

            // Right arrow button
            this.nextButton = new St.Button({
                reactive: true,
                can_focus: true,
                track_hover: true,
                style_class: 'navigation-button',
                width: 22,
                height: 22,
                child: new St.Icon({
                    icon_name: 'go-next-symbolic',
                    icon_size: 14,
                    style_class: 'navigation-icon',
                }),
            });

            navigationBox.add_child(this.prevButton);
            navigationBox.add_child(this.viewLabel);
            navigationBox.add_child(this.nextButton);
            navigationItem.actor.add_child(navigationBox);

            this.menu.addMenuItem(navigationItem);
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

            // Create containers for both views
            this.todayView = new St.BoxLayout({
                vertical: true,
                visible: true,
            });

            this.tomorrowView = new St.BoxLayout({
                vertical: true,
                visible: false,
            });

            // Create menu items for today's view
            this.currentPriceItem = new PopupMenu.PopupMenuItem(
                _('Current Price: Loading…')
            );
            this.nextPriceItem = new PopupMenu.PopupMenuItem(
                _('Next Price: Loading…')
            );
            this.lowestPriceItem = new PopupMenu.PopupMenuItem(
                _('Lowest Price Today: Loading…')
            );
            this.highestPriceItem = new PopupMenu.PopupMenuItem(
                _('Highest Price Today: Loading…')
            );
            this.averagePriceItem = new PopupMenu.PopupMenuItem(
                _('Average Price Today: Loading…')
            );
            this.cheapestHoursItem = new PopupMenu.PopupMenuItem(
                _('Cheapest Hours: Loading…')
            );
            this.expensiveHoursItem = new PopupMenu.PopupMenuItem(
                _('Most Expensive Hours: Loading…')
            );

            // Create menu items for tomorrow's view
            this.tomorrowLowestPriceItem = new PopupMenu.PopupMenuItem(
                _('Lowest Price Tomorrow: Loading…')
            );
            this.tomorrowHighestPriceItem = new PopupMenu.PopupMenuItem(
                _('Highest Price Tomorrow: Loading…')
            );
            this.tomorrowAveragePriceItem = new PopupMenu.PopupMenuItem(
                _('Average Price Tomorrow: Loading…')
            );
            this.tomorrowCheapestHoursItem = new PopupMenu.PopupMenuItem(
                _('Cheapest Hours Tomorrow: Loading…')
            );
            this.tomorrowExpensiveHoursItem = new PopupMenu.PopupMenuItem(
                _('Most Expensive Hours Tomorrow: Loading…')
            );

            // Add items to today's view
            const todayMenuContainer = new PopupMenu.PopupMenuSection();
            [
                this.currentPriceItem,
                this.nextPriceItem,
                new PopupMenu.PopupSeparatorMenuItem(),
                this.lowestPriceItem,
                this.highestPriceItem,
                this.averagePriceItem,
                this.cheapestHoursItem,
                this.expensiveHoursItem,
            ].forEach(item => todayMenuContainer.addMenuItem(item));

            // Add items to tomorrow's view
            const tomorrowMenuContainer = new PopupMenu.PopupMenuSection();
            [
                this.tomorrowLowestPriceItem,
                this.tomorrowHighestPriceItem,
                this.tomorrowAveragePriceItem,
                this.tomorrowCheapestHoursItem,
                this.tomorrowExpensiveHoursItem,
            ].forEach(item => tomorrowMenuContainer.addMenuItem(item));

            // Add containers to the menu
            const menuSection = new PopupMenu.PopupMenuSection();
            menuSection.actor.add_child(this.todayView);
            menuSection.actor.add_child(this.tomorrowView);
            this.menu.addMenuItem(menuSection);

            // Add the menu containers to their respective views
            this.todayView.add_child(todayMenuContainer.actor);
            this.tomorrowView.add_child(tomorrowMenuContainer.actor);

            // Add separator and common items
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

            // Add timestamp - only used for debugging, hide for now
            // this.timestampItem = new PopupMenu.PopupMenuItem(_('Last Updated: Never'));
            // this.menu.addMenuItem(this.timestampItem);

            // Add refresh button - only used for debugging, hide for now
            // const refreshButton = new PopupMenu.PopupMenuItem(_('Refresh'));
            // refreshButton.connect('activate', () => {
            //    this._refresh();
            // });
            // this.menu.addMenuItem(refreshButton);

            // Add attribution sections
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
            const attributionItem = new PopupMenu.PopupMenuItem('');
            const attributionBox = new St.BoxLayout({
                vertical: false,
                style_class: 'attribution-box',
            });

            const label = new St.Label({
                text: _('Price data: '),
                style_class: 'attribution-label',
                y_align: Clutter.ActorAlign.CENTER,
            });

            const link = new St.Button({
                reactive: true,
                track_hover: true,
                style_class: 'attribution-link',
                label: 'porssisahko.net',
                y_align: Clutter.ActorAlign.CENTER,
            });

            attributionBox.add_child(label);
            attributionBox.add_child(link);
            attributionItem.actor.add_child(attributionBox);
            this.menu.addMenuItem(attributionItem);

            // Connect navigation button signals
            this.prevButton.connect('clicked', () => this._switchView('today'));
            this.nextButton.connect('clicked', () =>
                this._switchView('tomorrow')
            );

            // Initial view setup
            this._updateNavigationState();

            // Store signal connections
            this._signalHandlers.push(
                this.prevButton.connect('clicked', () =>
                    this._switchView('today')
                )
            );
            this._signalHandlers.push(
                this.nextButton.connect('clicked', () =>
                    this._switchView('tomorrow')
                )
            );

            // Attribution link signal
            this._signalHandlers.push(
                link.connect('clicked', () => {
                    try {
                        Gio.app_info_launch_default_for_uri(
                            'https://porssisahko.net',
                            null
                        );
                        this.menu.close();
                    } catch (e) {
                        logError(e);
                    }
                })
            );
        }

        _switchView(view) {
            this._currentView = view;
            this.todayView.visible = view === 'today';
            this.tomorrowView.visible = view === 'tomorrow';

            const helsinkiTime = this._getHelsinkiTime();
            const date = new Date(helsinkiTime);
            if (view === 'tomorrow') date.setDate(date.getDate() + 1);

            const formattedDate = this._formatDate(date);

            this.viewLabel.set_text(
                view === 'today'
                    ? _('Today (%s)').format(formattedDate)
                    : _('Tomorrow (%s)').format(formattedDate)
            );

            this._updateNavigationState();
        }

        _updateNavigationState() {
            // Update navigation buttons state
            this.prevButton.reactive = this._currentView === 'tomorrow';
            this.nextButton.reactive = this._currentView === 'today';

            // Update button opacity to indicate state
            this.prevButton.opacity = this.prevButton.reactive ? 255 : 128;
            this.nextButton.opacity = this.nextButton.reactive ? 255 : 128;
        }

        _formatPrice(price) {
            // Format the number according to the current locale
            const formattedNumber = price.toLocaleString(undefined, {
                minimumFractionDigits: 2,
                maximumFractionDigits: 2,
            });

            // Translators: %s is the price value
            return _('%s c/kWh').format(formattedNumber);
        }

        _findPriceExtremes(prices, isTomorrow = false) {
            const helsinkiTime = this._getHelsinkiTime();
            const todayStart = new Date(helsinkiTime);
            todayStart.setHours(0, 0, 0, 0);

            // Create start and end dates based on whether we're looking at today or tomorrow
            const targetStart = new Date(todayStart);
            if (isTomorrow) targetStart.setDate(targetStart.getDate() + 1);

            const targetEnd = new Date(targetStart);
            targetEnd.setHours(23, 59, 59, 999);

            const targetPrices = prices.filter(price => {
                const priceDate = new Date(price.startDate);
                return priceDate >= targetStart && priceDate <= targetEnd;
            });

            if (targetPrices.length === 0) {
                return {
                    lowest: null,
                    highest: null,
                    cheapestHours: [],
                    expensiveHours: [],
                    averagePrice: 0,
                };
            }

            // Sort prices by price value
            const sortedByPrice = [...targetPrices].sort(
                (a, b) => a.price - b.price
            );

            const lowest = sortedByPrice[0];
            const highest = sortedByPrice[sortedByPrice.length - 1];

            const averagePrice =
                targetPrices.reduce((sum, price) => sum + price.price, 0) /
                targetPrices.length;

            // Function to find extreme hours without duplicates
            const findExtremes = (prices, isCheapest) => {
                const uniqueHours = new Set(); // Track unique hour periods
                const results = [];

                // Sort by price (ascending for cheapest, descending for most expensive)
                const sortedPrices = [...prices].sort((a, b) =>
                    isCheapest ? a.price - b.price : b.price - a.price
                );

                // Iterate through sorted prices until we have 3 unique periods
                for (const price of sortedPrices) {
                    const startHour = new Date(price.startDate).getHours();

                    // Check if we already have this hour
                    if (!uniqueHours.has(startHour) && results.length < 3) {
                        uniqueHours.add(startHour);
                        results.push({
                            startDate: new Date(price.startDate),
                            endDate: new Date(price.endDate),
                            price: price.price,
                        });
                    }

                    // Break if we have found 3 unique periods
                    if (results.length === 3) break;
                }

                // Return results already sorted by price (no need for additional sorting)
                return results;
            };

            const cheapestHours = findExtremes(targetPrices, true);
            const expensiveHours = findExtremes(targetPrices, false);

            return {
                lowest,
                highest,
                cheapestHours,
                expensiveHours,
                averagePrice,
            };
        }

        _updateUI(currentPrice, nextPrice, prices) {
            // Update panel display
            const displayPrice = this._formatPrice(currentPrice.price);
            this.priceLabel.set_text(displayPrice);

            // Get Helsinki time for date comparisons
            const helsinkiTime = this._getHelsinkiTime();
            const todayStart = new Date(helsinkiTime);
            todayStart.setHours(0, 0, 0, 0);
            const tomorrowStart = new Date(todayStart);
            tomorrowStart.setDate(tomorrowStart.getDate() + 1);
            const tomorrowEnd = new Date(tomorrowStart);
            tomorrowEnd.setHours(23, 59, 59, 999);

            // Split prices into today and tomorrow
            const todayPrices = prices.filter(price => {
                const priceDate = new Date(price.startDate);
                return priceDate >= todayStart && priceDate < tomorrowStart;
            });

            const tomorrowPrices = prices.filter(price => {
                const priceDate = new Date(price.startDate);
                return priceDate >= tomorrowStart && priceDate < tomorrowEnd;
            });

            // Check if we have complete data for tomorrow (should have 24 hours)
            // const hasTomorrowData = tomorrowPrices.length === 24;

            // Update current and next price
            this.currentPriceItem.label.set_text(
                _('Current Price: %s\nPeriod: %s - %s').format(
                    displayPrice,
                    this._formatDateTime(currentPrice.startDate),
                    this._formatDateTime(currentPrice.endDate)
                )
            );

            if (nextPrice) {
                const nextPriceFormatted = this._formatPrice(nextPrice.price);
                this.nextPriceItem.label.set_text(
                    _('Next Price: %s\nPeriod: %s - %s').format(
                        nextPriceFormatted,
                        this._formatDateTime(nextPrice.startDate),
                        this._formatDateTime(nextPrice.endDate)
                    )
                );
            } else {
                this.nextPriceItem.label.set_text(
                    _('Next Price: No data available')
                );
            }

            // Update today's data
            const todayExtremes = this._findPriceExtremes(todayPrices, false);
            this._updateDayPrices(todayExtremes, false);

            // Update tomorrow's data
            const tomorrowExtremes = this._findPriceExtremes(
                tomorrowPrices,
                true
            );
            this._updateDayPrices(tomorrowExtremes, true);

            // Update timestamp - only used for debugging, hide for now
            // this.timestampItem.label.set_text(_('Last Updated: %s').format(new Date().toLocaleString()));

            // Update icon based on current price
            if (currentPrice.price >= 15) {
                // Price is 15 cents or more
                this.icon.icon_name = 'dialog-warning-symbolic';
            } else if (currentPrice.price >= 10) {
                // Price is between 10-15 cents
                this.icon.icon_name = 'dialog-information-symbolic';
            } else {
                // Price is 10 cents or less
                this.icon.icon_name = ''; // Do not display icon
            }
        }

        _showNoTomorrowData() {
            const message = _(
                'Price data for tomorrow will be available after 14:20'
            );
            this.tomorrowLowestPriceItem.label.set_text(
                _('Lowest Price Tomorrow: %s').format(message)
            );
            this.tomorrowHighestPriceItem.label.set_text(
                _('Highest Price Tomorrow: %s').format(message)
            );
            this.tomorrowAveragePriceItem.label.set_text(
                _('Average Price Tomorrow: %s').format(message)
            );
            this.tomorrowCheapestHoursItem.label.set_text(
                _('Cheapest Hours Tomorrow: %s').format(message)
            );
            this.tomorrowExpensiveHoursItem.label.set_text(
                _('Most Expensive Hours Tomorrow: %s').format(message)
            );
        }

        _updateDayPrices(extremes, isTomorrow) {
            // const dayText = isTomorrow ? _('Tomorrow') : _('Today');

            if (extremes.lowest && extremes.highest) {
                // Update lowest and highest prices
                if (isTomorrow) {
                    this.tomorrowLowestPriceItem.label.set_text(
                        _('Lowest Price Tomorrow: %s\nTime: %s - %s').format(
                            this._formatPrice(extremes.lowest.price),
                            this._formatDateTime(extremes.lowest.startDate),
                            this._formatDateTime(extremes.lowest.endDate)
                        )
                    );

                    this.tomorrowHighestPriceItem.label.set_text(
                        _('Highest Price Tomorrow: %s\nTime: %s - %s').format(
                            this._formatPrice(extremes.highest.price),
                            this._formatDateTime(extremes.highest.startDate),
                            this._formatDateTime(extremes.highest.endDate)
                        )
                    );

                    this.tomorrowAveragePriceItem.label.set_text(
                        _('Average Price Tomorrow: %s').format(
                            this._formatPrice(extremes.averagePrice)
                        )
                    );
                } else {
                    this.lowestPriceItem.label.set_text(
                        _('Lowest Price Today: %s\nTime: %s - %s').format(
                            this._formatPrice(extremes.lowest.price),
                            this._formatDateTime(extremes.lowest.startDate),
                            this._formatDateTime(extremes.lowest.endDate)
                        )
                    );

                    this.highestPriceItem.label.set_text(
                        _('Highest Price Today: %s\nTime: %s - %s').format(
                            this._formatPrice(extremes.highest.price),
                            this._formatDateTime(extremes.highest.startDate),
                            this._formatDateTime(extremes.highest.endDate)
                        )
                    );

                    this.averagePriceItem.label.set_text(
                        _('Average Price Today: %s').format(
                            this._formatPrice(extremes.averagePrice)
                        )
                    );
                }

                // Update cheapest hours
                if (extremes.cheapestHours.length > 0) {
                    const cheapestHoursText = extremes.cheapestHours
                        .map(period =>
                            _('%s - %s (%s)').format(
                                this._formatDateTime(period.startDate),
                                this._formatDateTime(period.endDate),
                                this._formatPrice(period.price)
                            )
                        )
                        .join('\n');
                    if (isTomorrow) {
                        this.tomorrowCheapestHoursItem.label.set_text(
                            _('Cheapest Hours Tomorrow:\n%s').format(
                                cheapestHoursText
                            )
                        );
                    } else {
                        this.cheapestHoursItem.label.set_text(
                            _('Cheapest Hours Today:\n%s').format(
                                cheapestHoursText
                            )
                        );
                    }
                } else {
                    if (isTomorrow) {
                        this.tomorrowCheapestHoursItem.label.set_text(
                            _('Cheapest Hours Tomorrow: No data available')
                        );
                    } else {
                        this.cheapestHoursItem.label.set_text(
                            _('Cheapest Hours Today: No data available')
                        );
                    }
                }

                // Update most expensive hours
                if (extremes.expensiveHours.length > 0) {
                    const expensiveHoursText = extremes.expensiveHours
                        .map(period =>
                            _('%s - %s (%s)').format(
                                this._formatDateTime(period.startDate),
                                this._formatDateTime(period.endDate),
                                this._formatPrice(period.price)
                            )
                        )
                        .join('\n');

                    if (isTomorrow) {
                        this.tomorrowExpensiveHoursItem.label.set_text(
                            _('Most Expensive Hours Tomorrow:\n%s').format(
                                expensiveHoursText
                            )
                        );
                    } else {
                        this.expensiveHoursItem.label.set_text(
                            _('Most Expensive Hours Today:\n%s').format(
                                expensiveHoursText
                            )
                        );
                    }
                } else {
                    if (isTomorrow) {
                        this.tomorrowExpensiveHoursItem.label.set_text(
                            _(
                                'Most Expensive Hours Tomorrow: No data available'
                            )
                        );
                    } else {
                        this.expensiveHoursItem.label.set_text(
                            _('Most Expensive Hours Today: No data available')
                        );
                    }
                }
            } else {
                // Handle case when no data is available
                if (isTomorrow) {
                    this.tomorrowLowestPriceItem.label.set_text(
                        _('Lowest Price Tomorrow: No data available')
                    );
                    this.tomorrowHighestPriceItem.label.set_text(
                        _('Highest Price Tomorrow: No data available')
                    );
                    this.tomorrowAveragePriceItem.label.set_text(
                        _('Average Price Tomorrow: No data available')
                    );
                    this.tomorrowCheapestHoursItem.label.set_text(
                        _('Cheapest Hours Tomorrow: No data available')
                    );
                    this.tomorrowExpensiveHoursItem.label.set_text(
                        _('Most Expensive Hours Tomorrow: No data available')
                    );
                } else {
                    this.lowestPriceItem.label.set_text(
                        _('Lowest Price Today: No data available')
                    );
                    this.highestPriceItem.label.set_text(
                        _('Highest Price Today: No data available')
                    );
                    this.averagePriceItem.label.set_text(
                        _('Average Price Today: No data available')
                    );
                    this.cheapestHoursItem.label.set_text(
                        _('Cheapest Hours Today: No data available')
                    );
                    this.expensiveHoursItem.label.set_text(
                        _('Most Expensive Hours Today: No data available')
                    );
                }
            }
        }

        async _loadElectricityPrice() {
            try {
                // First, try to load from cache
                if (!this._cachedData)
                    this._cachedData = await this._loadCache();

                // Check if we need to fetch new data
                if (this._shouldFetchNewData()) {
                    try {
                        const prices = await this._fetchDataWithRetry();

                        // Save merged data to cache
                        await this._saveCache(prices);
                        this._cachedData = {
                            timestamp: this._getHelsinkiTime().toISOString(),
                            prices: [
                                ...(this._cachedData?.prices || []),
                                ...prices,
                            ],
                        };
                    } catch (error) {
                        // If fetching fails and we have cached data, use it
                        if (this._cachedData) {
                            logError(
                                'Using cached data due to fetch failure:',
                                error
                            );
                        } else {
                            throw error;
                        } // Re-throw if we have no cached data
                    }
                }

                const helsinkiTime = this._getHelsinkiTime();
                const sortedPrices = this._cachedData.prices.sort(
                    (a, b) => new Date(a.startDate) - new Date(b.startDate)
                );

                const currentPrice = sortedPrices.find(price => {
                    const startDate = new Date(price.startDate);
                    const endDate = new Date(price.endDate);
                    return startDate <= helsinkiTime && endDate > helsinkiTime;
                });

                const nextPrice = sortedPrices.find(
                    price => new Date(price.startDate) > helsinkiTime
                );

                if (!currentPrice)
                    throw new Error('No current price data available');

                return {currentPrice, nextPrice, prices: sortedPrices};
            } catch (error) {
                logError('Error in _loadElectricityPrice:', error);
                throw error;
            }
        }

        async _refresh() {
            try {
                const {currentPrice, nextPrice, prices} =
                    await this._loadElectricityPrice();
                this._updateUI(currentPrice, nextPrice, prices);
            } catch (error) {
                this.priceLabel.set_text(_('Error'));
                // Translators: %s is the error message
                this.currentPriceItem.label.set_text(
                    _('Error: %s').format(error.message)
                );
                this.nextPriceItem.label.set_text(
                    _('Next Price: Not available')
                );
                this.lowestPriceItem.label.set_text(
                    _('Lowest Price: Not available')
                );
                this.highestPriceItem.label.set_text(
                    _('Highest Price: Not available')
                );
                this.cheapestHoursItem.label.set_text(
                    _('Cheapest Hours: Not available')
                );
                this.expensiveHoursItem.label.set_text(
                    _('Cheapest Hours: Not available')
                );
                this.icon.icon_name = 'dialog-error-symbolic';
            }
        }

        destroy() {
            // if (this._refreshTimeout) GLib.source_remove(this._refreshTimeout);

            // Disconnect all signal handlers
            this._signalHandlers.forEach(handler => {
                if (handler) {
                    if (
                        this.prevButton &&
                        handler === this.prevButton.signalHandlers
                    ) {
                        this.prevButton.disconnect(handler);
                    } else if (
                        this.nextButton &&
                        handler === this.nextButton.signalHandlers
                    ) {
                        this.nextButton.disconnect(handler);
                    }
                    // For any other connections
                    else {
                        const [object, id] = handler;
                        if (object && id) object.disconnect(id);
                    }
                }
            });
            this._signalHandlers = [];

            // Remove timeout
            if (this._refreshTimeout) GLib.source_remove(this._refreshTimeout);

            super.destroy();
        }
    }
);

export default class Extension extends ExtensionBase {
    constructor(metadata) {
        super(metadata);
        // Initialize gettext for the extension
        ExtensionBase.prototype.initTranslations.call(this);
    }

    enable() {
        this._indicator = new ElectricityPriceIndicator();
        Main.panel.addToStatusArea('electricity-price', this._indicator);
    }

    disable() {
        if (this._indicator) {
            this._indicator.destroy();
            this._indicator = null;
        }
    }
}
